/* -LICENSE-START-
 ** Copyright (c) 2018 Blackmagic Design
 **
 ** Permission is hereby granted, free of charge, to any person or organization
 ** obtaining a copy of the software and accompanying documentation covered by
 ** this license (the "Software") to use, reproduce, display, distribute,
 ** execute, and transmit the Software, and to prepare derivative works of the
 ** Software, and to permit third-parties to whom the Software is furnished to
 ** do so, all subject to the following:
 **
 ** The copyright notices in the Software and this entire statement, including
 ** the above license grant, this restriction and the following disclaimer,
 ** must be included in all copies of the Software, in whole or in part, and
 ** all derivative works of the Software, unless such copies or derivative
 ** works are solely in the form of machine-executable object code generated by
 ** a source language processor.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 ** -LICENSE-END-
 */

import AppKit
import BluetoothControl
import Utility
import CCU
import CameraControlInterface

class GainSliderViewController: NSViewController {

	// IBOutlets
	@IBOutlet weak var valueView: NSTextField?
	@IBOutlet weak var gainSlider: BlackmagicSlider?


	// Member variables
	weak var cciOut: OutgoingCameraControlFromUIDelegate?

	var valueIndex = 0
	var gainValues = VideoConfig.gainValues[CameraModel.Unknown]!
	var outputGainMoment: [Decibels:Date] = [:]

	
	//==================================================
	//	UIViewController methods
	//==================================================
	override func viewDidLoad() {
		super.viewDidLoad()

		let appDelegate = NSApplication.shared.delegate as! AppDelegate
		let cameraControlInterface: CameraControlInterface = appDelegate.cameraControlInterface
		cciOut = cameraControlInterface
	}

	override func viewDidAppear() {
		super.viewDidAppear()

		if let gainSlider = gainSlider {
			gainSlider.minValue = Double(0)
			gainSlider.maxValue = Double(gainValues.count - 1)
			gainSlider.setCallbacks(onTentativeValueChanged: onGainSliderChanged, onValueChanged: onGainSliderSet)
		}

		updateViews()
	}

	override func viewDidDisappear() {
		super.viewDidDisappear()

		outputGainMoment.removeAll()
	}

	@IBAction func onDecreaseButtonClicked(_: NSView) {
		valueIndex = valueIndex > 0 ? valueIndex - 1 : gainValues.count - 1

		updateViews()
		outputGain()
	}

	@IBAction func onIncreaseButtonClicked(_: NSView) {
		valueIndex = valueIndex < gainValues.count - 1 ? valueIndex + 1 : 0

		updateViews()
		outputGain()
	}

	//==================================================
	//	BlackmagicSlider callbacks
	//==================================================
	func onGainSliderChanged(slider: BlackmagicSlider) {
		let index = Int(slider.floatValue.rounded())
		if 0...(gainValues.count - 1) ~= index {
			valueIndex = index
			updateTextView()
		}
	}

	func onGainSliderSet(slider: BlackmagicSlider) {
		let index = Int(slider.floatValue.rounded())
		if 0...(gainValues.count - 1) ~= index {
			valueIndex = index
			updateTextView()
			outputGain()
		}
	}

	//==================================================
	//	IncomingCameraControlToUIDelegate methods
	//==================================================
	func onCameraSpecificationReceived(_ cameraModel: Utility.CameraModel) {
		gainValues = VideoConfig.gainValues[cameraModel] ?? VideoConfig.gainValues[CameraModel.Unknown]!

		if let gainSlider = gainSlider {
			gainSlider.minValue = Double(0)
			gainSlider.maxValue = Double(gainValues.count - 1)
		}
	}

	func outputGain() {
		let gain = gainValues[valueIndex]
		outputGainMoment[gain] = Date()
		cciOut?.onGainSet(gain)
	}

	func onGainReceived(_ gain: Decibels) {
		var accept = true
		if let moment = outputGainMoment.removeValue(forKey: gain) {
			let delta: TimeInterval = Date().timeIntervalSince(moment)
			accept = delta >= 0.25
		}

		if accept {
			if let index = gainValues.firstIndex(of: gain) {
				valueIndex = index
				updateViews()
			}
		}
	}
	
	//==================================================
	//	UI control
	//==================================================	
	func clearCallbacks() {
		gainSlider?.clearCallbacks()
	}

	func updateViews() {
		updateSliderView()
		updateTextView()
	}

	func updateSliderView() {
		if let gainSlider = gainSlider {
			gainSlider.setNewValue(Float(valueIndex))
		}
	}

	func updateTextView() {
		if let valueView = valueView {
			valueView.stringValue = "\(gainValues[valueIndex])"
		}
	}
}
