/* -LICENSE-START-
 ** Copyright (c) 2018 Blackmagic Design
 **
 ** Permission is hereby granted, free of charge, to any person or organization
 ** obtaining a copy of the software and accompanying documentation covered by
 ** this license (the "Software") to use, reproduce, display, distribute,
 ** execute, and transmit the Software, and to prepare derivative works of the
 ** Software, and to permit third-parties to whom the Software is furnished to
 ** do so, all subject to the following:
 **
 ** The copyright notices in the Software and this entire statement, including
 ** the above license grant, this restriction and the following disclaimer,
 ** must be included in all copies of the Software, in whole or in part, and
 ** all derivative works of the Software, unless such copies or derivative
 ** works are solely in the form of machine-executable object code generated by
 ** a source language processor.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 ** -LICENSE-END-
 */

import AppKit
import BluetoothControl
import Utility
import CCU
import CameraControlInterface

class MainViewController: NSViewController, IncomingCameraControlToUIDelegate {

	// IBOutlets
    @IBOutlet weak var whiteBalanceLabel: NSTextField!
    @IBOutlet weak var whiteBalanceSlider: BlackmagicSlider!

    @IBOutlet weak var tintLabel: NSTextField!
    @IBOutlet weak var tintSlider: BlackmagicSlider!

    @IBOutlet weak var offSpeedFrameRateLabel: NSTextField!
    @IBOutlet weak var offSpeedFrameRateSlider: BlackmagicSlider!
    @IBOutlet weak var offSpeedFrameRatePrevButton: NSButton!
    @IBOutlet weak var offSpeedFrameRateNextButton: NSButton!
    @IBOutlet weak var offSpeedFrameRateToggleButton: NSButton!

    @IBOutlet weak var apertureLabel: NSTextField!
    @IBOutlet weak var apertureSlider: BlackmagicSlider!
	@IBOutlet weak var aperturePrevButton: NSButton!
	@IBOutlet weak var apertureNextButton: NSButton!

    @IBOutlet weak var shutterLabel: NSTextField!
    @IBOutlet weak var shutterSlider: BlackmagicSlider!

    @IBOutlet weak var preset1Button: NSButton!
    @IBOutlet weak var preset2Button: NSButton!
    @IBOutlet weak var preset3Button: NSButton!
    @IBOutlet weak var preset4Button: NSButton!
    @IBOutlet weak var preset5Button: NSButton!
    @IBOutlet weak var autoWhiteBalanceButton: NSButton!

	@IBOutlet weak var gainSliderContainer: NSView!
	weak var gainSliderController: GainSliderViewController?
	@IBOutlet weak var isoButtonContainer: NSView!
	weak var isoButtonController: ISOButtonViewController?

   	@IBOutlet weak var ndFilterView: NSBox!
	@IBOutlet weak var ndFilterLabel: NSTextField!

	@IBOutlet weak var focusPullLabel: NSTextField!
	@IBOutlet weak var focusPullSlider: BlackmagicSlider!

	// Member variables
	weak var m_outgoingCameraControlDelegate: OutgoingCameraControlFromUIDelegate?
	var m_shutterValueIsAngle: Bool = true
	var presetButtons = [NSButton]()
    var m_canChangeNDFilter = true
	
	enum ISOControlType {
//		case GainButtons
		case GainSlider
		case ISOButtons
//		case ISOSlider
	}

	var isoControlType = ISOControlType.ISOButtons

	func getISOControlType(of cameraModel : CameraModel) -> ISOControlType {
		switch cameraModel
		{
		case	.URSABroadcast,
				.URSABroadcastG2:			return ISOControlType.GainSlider
		default:							return ISOControlType.ISOButtons
		}
	}

	//==================================================
	//	UIViewController methods
	//==================================================
	override func viewDidLoad() {
        super.viewDidLoad()

        presetButtons.append(preset1Button)
        presetButtons.append(preset2Button)
        presetButtons.append(preset3Button)
        presetButtons.append(preset4Button)
        presetButtons.append(preset5Button)
        presetButtons.append(autoWhiteBalanceButton)

        whiteBalanceSlider.minValue = Double(VideoConfig.kWhiteBalanceMin)
        whiteBalanceSlider.maxValue = Double(VideoConfig.kWhiteBalanceMax)
		whiteBalanceSlider.setCallbacks(onTentativeValueChanged: onWhiteBalanceSliderChanged, onValueChanged: onWhiteBalanceSliderSet)

		tintSlider.minValue = Double(VideoConfig.kTintMin)
        tintSlider.maxValue = Double(VideoConfig.kTintMax)
		tintSlider.setCallbacks(onTentativeValueChanged: onTintSliderChanged, onValueChanged: onTintSliderSet)

		apertureSlider.minValue = 0.0
        apertureSlider.maxValue = 1.0
		apertureSlider.setCallbacks(onTentativeValueChanged: nil, onValueChanged: onApertureSliderSet)

		shutterSlider.minValue = 5.0
        shutterSlider.maxValue = 360.0
		shutterSlider.setCallbacks(onTentativeValueChanged: onShutterSliderChanged, onValueChanged: onShutterSliderSet)

		offSpeedFrameRateSlider.minValue = Double(VideoConfig.kOffSpeedFrameRateMin[CameraModel.Unknown]!)
        offSpeedFrameRateSlider.maxValue = Double(VideoConfig.kOffSpeedFrameRateMax[CameraModel.Unknown]!)
        offSpeedFrameRateSlider.addImportantValues([50.0, 40.0, 30.0])
		offSpeedFrameRateSlider.setCallbacks(onTentativeValueChanged: onOffSpeedFrameRateSliderChanged, onValueChanged: onOffSpeedFrameRateSliderSet, onImportantValue: onOffSpeedFrameRateSliderImportantValue)

		focusPullSlider.minValue = -100
		focusPullSlider.maxValue = 100
		focusPullSlider.setCallbacks(onTentativeValueChanged: onFocusPullSliderChanged, onValueChanged: onFocusPullSliderSet)

		let appDelegate = NSApplication.shared.delegate as! AppDelegate
		let cameraControlInterface: CameraControlInterface = appDelegate.cameraControlInterface
		m_outgoingCameraControlDelegate = cameraControlInterface

		let cameraModel = cameraControlInterface.getCameraModel()
		updateISOControl(cameraModel)

		// Register this view controller as a CameraControlToUIDelegate with the CameraControlInterface,
		// and assign CameraControlInterface as our OutgoingCameraControlFromUIDelegate.
        cameraControlInterface.m_cameraControlToUIDelegate = self
        cameraControlInterface.onControlViewLoaded()
    }

	override func prepare(for segue: NSStoryboardSegue, sender: Any?) {
		// Store references to view controllers for switching, and animating transitions.
		if let vc = segue.destinationController as? GainSliderViewController {
			gainSliderController = vc
		}
		else if let vc = segue.destinationController as? ISOButtonViewController {
			isoButtonController = vc
		}
	}

	//==================================================
	//	IBActions
	//==================================================
    @IBAction func onPrevWhiteBalanceButtonClicked(_: NSButton) {
        if let newValues: (whiteBalance: Int16, presetIndex: Int) = m_outgoingCameraControlDelegate?.onWhiteBalanceDecremented() {
            updateWhiteBalanceWidgets(newValues.whiteBalance)
            presetButtons.forEach { $0.isSelected = $0.tag == newValues.presetIndex }
        }
    }

    @IBAction func onNextWhiteBalanceButtonClicked(_: NSButton) {
        if let newValues: (whiteBalance: Int16, presetIndex: Int) = m_outgoingCameraControlDelegate?.onWhiteBalanceIncremented() {
            updateWhiteBalanceWidgets(newValues.whiteBalance)
            presetButtons.forEach { $0.isSelected = $0.tag == newValues.presetIndex }
        }
    }

    @IBAction func onPrevTintButtonClicked(_: NSButton) {
        if let newValues: (tint: Int16, presetIndex: Int) = m_outgoingCameraControlDelegate?.onTintDecremented() {
            updateTintWidgets(newValues.tint)
            presetButtons.forEach { $0.isSelected = $0.tag == newValues.presetIndex }
        }
    }

    @IBAction func onNextTintButtonClicked(_: NSButton) {
        if let newValues: (tint: Int16, presetIndex: Int) = m_outgoingCameraControlDelegate?.onTintIncremented() {
            updateTintWidgets(newValues.tint)
            presetButtons.forEach { $0.isSelected = $0.tag == newValues.presetIndex }
        }
    }

    @IBAction func onPrevOffSpeedFrameRateButtonClicked(_: NSButton) {
        if let newOffSpeedFrameRate = m_outgoingCameraControlDelegate?.onOffSpeedFrameRateDecremented() {
            updateOffSpeedFrameRateWidgets(newOffSpeedFrameRate, true)
        }
    }

    @IBAction func onNextOffSpeedFrameRateButtonClicked(_: NSButton) {
        if let newOffSpeedFrameRate = m_outgoingCameraControlDelegate?.onOffSpeedFrameRateIncremented() {
            updateOffSpeedFrameRateWidgets(newOffSpeedFrameRate, true)
        }
    }

    @IBAction func offSpeedButtonTapped(_ sender: NSButton) {
        m_outgoingCameraControlDelegate?.onOffSpeedFrameRateToggled(sender.isSelected)
    }

	@IBAction func onPrevApertureButtonClicked(_: NSButton) {
		if let outgoingCameraControlDelegate = m_outgoingCameraControlDelegate {
			outgoingCameraControlDelegate.onApertureDecremented(apertureSlider.floatValue)
		}
	}

    @IBAction func onNextApertureButtonClicked(_: NSButton) {
		if let outgoingCameraControlDelegate = m_outgoingCameraControlDelegate {
			outgoingCameraControlDelegate.onApertureIncremented(apertureSlider.floatValue)
		}
	}

    @IBAction func onPrevShutterButtonClicked(_: NSButton) {
        let newShutterValue = m_outgoingCameraControlDelegate?.onShutterDecremented()
        if newShutterValue != nil {
            updateShutterWidgets(Float(newShutterValue!))
        }
    }

    @IBAction func onNextShutterButtonClicked(_: NSButton) {
        let newShutterValue = m_outgoingCameraControlDelegate?.onShutterIncremented()
        if newShutterValue != nil {
            updateShutterWidgets(Float(newShutterValue!))
        }
    }

    @IBAction func onPresetButtonClicked(_ sender: NSButton) {
        let currentlySelected = !sender.isSelected
        let selectedButtons = presetButtons.filter { return $0.isSelected }
        let saveCustomValues = selectedButtons.count != 0
        presetButtons.forEach { $0.isSelected = false }

        let presetIndex = sender.tag
        let newValues: (whiteBalance: Int16, tint: Int16)? = m_outgoingCameraControlDelegate?.onWhiteBalancePresetPressed(presetIndex, currentlySelected, saveCustomValues)
        if newValues != nil {
            updateWhiteBalanceWidgets(newValues!.whiteBalance)
            updateTintWidgets(newValues!.tint)
        }

        sender.isSelected = !currentlySelected
        presetButtons.last!.isSelected = false
    }
	
	@IBAction func onPrevFocusPullButtonClicked(_: NSButton) {
		if let focusPullStep = m_outgoingCameraControlDelegate?.onFocusPullDecremented() {
			updateFocusPullWidgets(focusPullStep)
		}
	}
	
	@IBAction func onNextFocusPullButtonClicked(_: NSButton) {
		if let focusPullStep = m_outgoingCameraControlDelegate?.onFocusPullIncremented() {
			updateFocusPullWidgets(focusPullStep)
		}
	}

	fileprivate var ndFilterStops: Float = 0.0 {
		didSet {
			updateNDFilterLabel()
		}
	}

	@IBAction func onPrevNDFilterButtonClicked(_ sender: NSButton) {
		if m_canChangeNDFilter {
			guard let stops = m_outgoingCameraControlDelegate?.onNDFilterDecremented() else { return }
			ndFilterStops = stops
        }
	}
	
	@IBAction func onNextNDFilterButtonClicked(_ sender: NSButton) {
		if m_canChangeNDFilter {
			guard let stops = m_outgoingCameraControlDelegate?.onNDFilterIncremented() else { return }
			ndFilterStops = stops
		}
	}

	//==================================================
	//	BlackmagicSlider callbacks
	//==================================================
    func onWhiteBalanceSliderChanged(_: BlackmagicSlider) {
        let newWhiteBalance = (Int16(whiteBalanceSlider.floatValue) / VideoConfig.kWhiteBalanceStep) * VideoConfig.kWhiteBalanceStep
        updateWhiteBalanceWidgets(newWhiteBalance)
    }

    func onWhiteBalanceSliderSet(_: BlackmagicSlider) {
        let newWhiteBalance = (Int16(whiteBalanceSlider.floatValue) / VideoConfig.kWhiteBalanceStep) * VideoConfig.kWhiteBalanceStep
        let presetIndex: Int? = m_outgoingCameraControlDelegate?.onWhiteBalanceChanged(newWhiteBalance)
        presetButtons.forEach { $0.isSelected = $0.tag == presetIndex }
    }

    func onTintSliderChanged(_: BlackmagicSlider) {
        let newTint = Int16(tintSlider.floatValue)
        updateTintWidgets(newTint)
    }

    func onTintSliderSet(_: BlackmagicSlider) {
        let newTint = Int16(tintSlider.floatValue)
        let presetIndex: Int? = m_outgoingCameraControlDelegate?.onTintChanged(newTint)
        presetButtons.forEach { $0.isSelected = $0.tag == presetIndex }
    }

    func onApertureSliderSet(_: BlackmagicSlider) {
		if let outgoingCameraControlDelegate = m_outgoingCameraControlDelegate {
			let apertureNormalised = apertureSlider.floatValue
			outgoingCameraControlDelegate.onApertureNormalisedChanged(apertureNormalised)
		}
    }

    func onShutterSliderChanged(_: BlackmagicSlider) {
        let newShutterValue = shutterSlider.floatValue.rounded()
        updateShutterWidgets(newShutterValue)
    }

    func onShutterSliderSet(_: BlackmagicSlider) {
        let newShutterValue = shutterSlider.floatValue.rounded()
        updateShutterWidgets(newShutterValue)
        m_outgoingCameraControlDelegate?.onShutterChanged(Double(newShutterValue))
    }

    func onOffSpeedFrameRateSliderChanged(_: BlackmagicSlider) {
        let newOffSpeedFrameRate = Int16(offSpeedFrameRateSlider.floatValue)
        offSpeedFrameRateLabel.stringValue = "\(newOffSpeedFrameRate) \(String.Localized("FrameRate.FPS"))"
    }

    func onOffSpeedFrameRateSliderSet(_: BlackmagicSlider) {
        let newOffSpeedFrameRate = Int16(offSpeedFrameRateSlider.floatValue)
        m_outgoingCameraControlDelegate?.onOffSpeedFrameRateChanged(newOffSpeedFrameRate)
        offSpeedFrameRateLabel.stringValue = "\(newOffSpeedFrameRate) \(String.Localized("FrameRate.FPS"))"
    }

    func onOffSpeedFrameRateSliderImportantValue(_ value: Double) {
        let newOffSpeedFrameRate = Int16(value)
        m_outgoingCameraControlDelegate?.onOffSpeedFrameRateChanged(newOffSpeedFrameRate)
        offSpeedFrameRateLabel.stringValue = "\(newOffSpeedFrameRate) \(String.Localized("FrameRate.FPS"))"
    }
	
	func onFocusPullSliderChanged(_: BlackmagicSlider) {
		let focusPullStep = Int(focusPullSlider.floatValue)
		m_outgoingCameraControlDelegate?.onFocusPullChanged(focusPullStep)
		updateFocusPullWidgets(focusPullStep)
	}
	
	func onFocusPullSliderSet(_: BlackmagicSlider) {
		// Special: Snap back to 0 once slider knob is released
		let focusPullStep = Int(focusPullSlider.floatValue)
		m_outgoingCameraControlDelegate?.onFocusPullChanged(focusPullStep)
		updateFocusPullWidgets(focusPullStep)
//		updateFocusPullWidgets(0)
	}

	//==================================================
	//	IncomingCameraControlToUIDelegate methods
	//==================================================
    func onWhiteBalanceReceived(_ whiteBalance: Int16, _ tint: Int16, _ presetIndex: Int) {
        updateWhiteBalanceWidgets(whiteBalance)
        updateTintWidgets(tint)
        presetButtons.forEach { $0.isSelected = $0.tag == presetIndex }
    }

    func onOffSpeedFrameRateReceived(_ offSpeedEnabled: Bool, _ frameRate: Int16) {
        updateOffSpeedFrameRateWidgets(frameRate, offSpeedEnabled)
    }

	func onApertureFstopReceived(_ fstop: Float, _ stopUnits: LensConfig.ApertureUnits) {
        apertureLabel.stringValue = LensConfig.GetStopString(fstop, stopUnits)
	}

	func onApertureNormalisedReceived(_ apertureNormalised: Float) {
		let valid = LensConfig.IsApertureNormalisedValid(apertureNormalised)
		apertureSlider.floatValue = valid ? apertureNormalised : 0.0
	}

//    func onShutterReceived(_ shutterAngle: Double) {
//        updateShutterWidgets(Float(shutterAngle))
//    }
	
	func onShutterAngleReceived(_ shutterAngle: Double) {
		m_shutterValueIsAngle = true
		shutterSlider.minValue = VideoConfig.kShutterAngleMin
		shutterSlider.maxValue = VideoConfig.kShutterAngleMax
		updateShutterWidgets(Float(shutterAngle))
	}
	
	func onShutterSpeedReceived(_ shutterSpeed: Int32) {
		m_shutterValueIsAngle = false
		shutterSlider.minValue = Double(VideoConfig.kShutterSpeedMin)
		shutterSlider.maxValue = Double(VideoConfig.kShutterSpeedMax)
		updateShutterWidgets(Float(shutterSpeed))
	}

	func onISOReceived(_ isoIndex: Int) {
		isoButtonController?.onISOReceived(isoIndex)
	}
	
	func onGainReceived(_ gain: Decibels) {
		gainSliderController?.onGainReceived(gain)
	}
	
	func setIrisControlEnabled(_ enabled: Bool)
	{
		apertureSlider.isEnabled = enabled
		aperturePrevButton.isEnabled = enabled
		apertureNextButton.isEnabled = enabled
	}

	func setShutterControlEnabled(_ enabled: Bool) {}
	func onTransportModeReceived(_ transportMode: CCUPacketTypes.MediaTransportMode) {}
	
	fileprivate var ndFilterDisplayMode = VideoConfig.NDFilterDisplayMode.Stop {
		didSet {
			updateNDFilterLabel()
		}
	}

	fileprivate func updateNDFilterLabel() {
		guard let labels = VideoConfig.kNDFilterTextForStopForDisplayMode[ndFilterStops] else { return }
		guard let label = labels[ndFilterDisplayMode] else { return }
		ndFilterLabel.stringValue = label
	}

	func onNDFilterStopReceived(_ stops: Float, _ mode: VideoConfig.NDFilterDisplayMode)
	{
		ndFilterStops = stops
		ndFilterDisplayMode = mode
	}

    func onCameraSpecificationReceived(_ cameraModel: CameraModel) {
		updateISOControl(cameraModel)

		ndFilterView.isHidden = !VideoConfig.kProductHasNDFilter.contains(cameraModel)
		m_canChangeNDFilter = VideoConfig.kProductCanChangeNDFilterRemotely.contains(cameraModel)
    }

	//==================================================
	//	UI control
	//==================================================
	func updateWhiteBalanceWidgets(_ whiteBalance: Int16) {
		whiteBalanceLabel.stringValue = "\(whiteBalance)K"
		whiteBalanceSlider.floatValue = Float(whiteBalance)
	}
	
	func updateTintWidgets(_ tint: Int16) {
		tintLabel.stringValue = "\(tint)"
		tintSlider.floatValue = Float(tint)
	}
	
	func updateOffSpeedFrameRateWidgets(_ frameRate: Int16, _ offSpeedEnabled: Bool) {
		offSpeedFrameRateLabel.stringValue = offSpeedEnabled ? "\(frameRate) \(String.Localized("FrameRate.FPS"))" : String.Localized("FrameRate.Sync")
		if offSpeedEnabled {
			offSpeedFrameRateSlider.floatValue = Float(frameRate)
		}
		
		offSpeedFrameRateSlider.isEnabled = offSpeedEnabled
		offSpeedFrameRateNextButton.isEnabled = offSpeedEnabled
		offSpeedFrameRatePrevButton.isEnabled = offSpeedEnabled
		offSpeedFrameRateToggleButton.isSelected = offSpeedEnabled
	}
	
	func updateShutterWidgets(_ newShutterValue: Float) {
		shutterSlider.floatValue = newShutterValue
		
		if (m_shutterValueIsAngle) {
			shutterLabel.stringValue = "\(newShutterValue.getStringValue())°"
		}
		else {
			shutterLabel.stringValue = "1/\(Int(newShutterValue))"
		}
	}
	
	func updateFocusPullWidgets(_ focusPullStep: Int) {
		if (focusPullStep == 0) {
			focusPullLabel.stringValue = "Idle"
		}
		else if (focusPullStep > 0) {
			focusPullLabel.stringValue = "+ \(focusPullStep)"
		}
		else {
			focusPullLabel.stringValue = "\(focusPullStep)"
		}
		
		focusPullSlider.floatValue = Float(focusPullStep)
	}

	func updateISOControl(_ cameraModel: CameraModel) {
		isoControlType = getISOControlType(of: cameraModel)

		gainSliderContainer.isHidden = true
		isoButtonContainer.isHidden = true
		switch isoControlType
		{
		case .GainSlider:
			gainSliderContainer.isHidden = false
			gainSliderController?.onCameraSpecificationReceived(cameraModel)
		case .ISOButtons:
			isoButtonContainer.isHidden = false
			isoButtonController?.onCameraSpecificationReceived(cameraModel)
		}
	}
}
