/* -LICENSE-START-
 ** Copyright (c) 2018 Blackmagic Design
 **
 ** Permission is hereby granted, free of charge, to any person or organization
 ** obtaining a copy of the software and accompanying documentation covered by
 ** this license (the "Software") to use, reproduce, display, distribute,
 ** execute, and transmit the Software, and to prepare derivative works of the
 ** Software, and to permit third-parties to whom the Software is furnished to
 ** do so, all subject to the following:
 **
 ** The copyright notices in the Software and this entire statement, including
 ** the above license grant, this restriction and the following disclaimer,
 ** must be included in all copies of the Software, in whole or in part, and
 ** all derivative works of the Software, unless such copies or derivative
 ** works are solely in the form of machine-executable object code generated by
 ** a source language processor.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 ** -LICENSE-END-
 */

import AppKit
import BluetoothControl
import Utility
import CameraControlInterface

class NonInteractiveTextField: NSTextField {
	override func hitTest(_ point: NSPoint) -> NSTextField? {
		return nil
	}
}

class TransportViewController: NSViewController, IncomingRecordControlToUIDelegate {

	// IBOutlets
	@IBOutlet weak var recordButton: NSButton!
	@IBOutlet weak var recordButtonLabel: NSTextField!
	@IBOutlet weak var stillCaptureButton: NSButton!
	@IBOutlet weak var stillCaptureButtonLabel: NSTextField!
	@IBOutlet weak var recordingErrorLabel: NonInteractiveTextField!
	@IBOutlet weak var timecodeLabel: NSTextField!
	@IBOutlet weak var slot0RemainingTimeLabel: NSTextField!
	@IBOutlet weak var slot0Box: NSBox!
	@IBOutlet weak var slot1RemainingTimeLabel: NSTextField!
	@IBOutlet weak var slot1Box: NSBox!
	@IBOutlet weak var slot2RemainingTimeLabel: NSTextField!
	@IBOutlet weak var slot2Box: NSBox!

	// Member variables
	weak var m_outgoingRecordControlDelegate: OutgoingRecordControlFromUIDelegate?

	//==================================================
	//	UIViewController methods
	//==================================================
	override func viewWillAppear() {
		let appDelegate = NSApplication.shared.delegate as! AppDelegate
		let connectedAsBluetooth = appDelegate.currentTransportBluetooth
		
		if (connectedAsBluetooth) {
			recordButton.isHidden = false
			recordButtonLabel.isHidden = false
			recordingErrorLabel.isHidden = true
			stillCaptureButton.isHidden = true
			stillCaptureButtonLabel.isHidden = true
			timecodeLabel.isHidden = false
			slot0Box.isHidden = false
			slot0RemainingTimeLabel.isHidden = false
			slot1Box.isHidden = false
			slot1RemainingTimeLabel.isHidden = false
			slot2Box.isHidden = false
			slot2RemainingTimeLabel.isHidden = false
		}
		else {
			recordButton.isHidden = false
			recordButtonLabel.isHidden = false
			recordingErrorLabel.isHidden = true
			stillCaptureButton.isHidden = false
			stillCaptureButtonLabel.isHidden = false
			timecodeLabel.isHidden = true
			slot0Box.isHidden = true
			slot0RemainingTimeLabel.isHidden = true
			slot1Box.isHidden = true
			slot1RemainingTimeLabel.isHidden = true
			slot2Box.isHidden = true
			slot2RemainingTimeLabel.isHidden = true
		}
	}
	
    override func viewDidLoad() {
		// Register as a IncomingRecordControlToUIDelegate with the CameraControlInterface,
		// and assign CameraControlInterface as our OutgoingRecordControlFromUIDelegate.
        let appDelegate = NSApplication.shared.delegate as! AppDelegate
        let cameraControlInterface: CameraControlInterface = appDelegate.cameraControlInterface
        m_outgoingRecordControlDelegate = cameraControlInterface
        cameraControlInterface.m_recordControlToUIDelegate = self
        cameraControlInterface.onTransportViewLoaded()

        super.viewDidLoad()
    }

	//==================================================
	//	IBActions
	//==================================================
    @IBAction func recordButtonTapped(_: NSButton) {
        m_outgoingRecordControlDelegate?.onRecordPressed()
    }
	
	@IBAction func stillCaptureButtonTapped(_: NSButton) {
		m_outgoingRecordControlDelegate?.onStillCapturePressed()
	}

	@IBAction func slotButtonTapped(button: NSButton) {
		m_outgoingRecordControlDelegate?.onSlotPressed(button.tag)
	}

	//==================================================
    //	IncomingRecordControlToUserDelegate methods
	//==================================================
    func setRecording(_ isRecording: Bool) {
        let colorFunc = isRecording ? showRecordingColors : showPreviewColors
        colorFunc()
    }

	func onActiveDisksReceived(_ active: [Bool]) {
		let activeSuffix = " \(String.Localized("Transport.Active"))"
		slot0Box.title = String.Localized("Transport.Slot0") + (active.count >= 1 && active[0] ? activeSuffix : "")
		slot1Box.title = String.Localized("Transport.Slot1") + (active.count >= 2 && active[1] ? activeSuffix : "")
		slot2Box.title = String.Localized("Transport.Slot2") + (active.count >= 3 && active[2] ? activeSuffix : "")
	}

    func onTimecodeReceived(_ timecode: String) {
        timecodeLabel.stringValue = timecode
    }

	func onRecordTimeRemainingReceived(_ remainingTimes: [String]) {
		if remainingTimes.count >= 1 {
			slot0RemainingTimeLabel.stringValue = remainingTimes[0]
		}
		if remainingTimes.count >= 2 {
			slot1RemainingTimeLabel.stringValue = remainingTimes[1]
		}
		if remainingTimes.count >= 3 {
			slot2RemainingTimeLabel.stringValue = remainingTimes[2]
		}
	}

    func setRecordingError(_ hasRecordingError: Bool) {
        recordingErrorLabel.isHidden = !hasRecordingError
    }
    
    func recordingEnabled(_ enabled: Bool) {
        recordButton.isEnabled = enabled
    }
    
    func setRecordTimeWarning(_ recordTimeWarning: RecordTimeWarning, _ cardIndex: Int) {}
	func onTimelapseReceived(_ isTimelapseActive: Bool) {}
	func showPlaybackView(for clipName: String, at playbackSpeed: Int) {}
	func hidePlaybackView() {}
	func updateClipName(_ clipName: String) {}

	//==================================================
	//	UI Control
	//==================================================
    func showRecordingColors() {
        let style = NSMutableParagraphStyle()
        style.alignment = .center
        let font = NSFont.systemFont(ofSize: 66)
        recordButton?.attributedTitle = NSAttributedString(string: "●", attributes: convertToOptionalNSAttributedStringKeyDictionary([convertFromNSAttributedStringKey(NSAttributedString.Key.foregroundColor): NSColor.red, convertFromNSAttributedStringKey(NSAttributedString.Key.paragraphStyle): style, convertFromNSAttributedStringKey(NSAttributedString.Key.font): font]))
    }

    func showPreviewColors() {
        let style = NSMutableParagraphStyle()
        style.alignment = .center
        let font = NSFont.systemFont(ofSize: 66)
        recordButton?.attributedTitle = NSAttributedString(string: "●", attributes: convertToOptionalNSAttributedStringKeyDictionary([convertFromNSAttributedStringKey(NSAttributedString.Key.foregroundColor): NSColor.lightGray, convertFromNSAttributedStringKey(NSAttributedString.Key.paragraphStyle): style, convertFromNSAttributedStringKey(NSAttributedString.Key.font): font]))
    }
}

// Helper function inserted by Swift 4.2 migrator.
fileprivate func convertToOptionalNSAttributedStringKeyDictionary(_ input: [String: Any]?) -> [NSAttributedString.Key: Any]? {
	guard let input = input else { return nil }
	return Dictionary(uniqueKeysWithValues: input.map { key, value in (NSAttributedString.Key(rawValue: key), value)})
}

// Helper function inserted by Swift 4.2 migrator.
fileprivate func convertFromNSAttributedStringKey(_ input: NSAttributedString.Key) -> String {
	return input.rawValue
}
