/* -LICENSE-START-
** Copyright (c) 2020 Blackmagic Design
**
** Permission is hereby granted, free of charge, to any person or organization
** obtaining a copy of the software and accompanying documentation covered by
** this license (the "Software") to use, reproduce, display, distribute,
** execute, and transmit the Software, and to prepare derivative works of the
** Software, and to permit third-parties to whom the Software is furnished to
** do so, all subject to the following:
**
** The copyright notices in the Software and this entire statement, including
** the above license grant, this restriction and the following disclaimer,
** must be included in all copies of the Software, in whole or in part, and
** all derivative works of the Software, unless such copies or derivative
** works are solely in the form of machine-executable object code generated by
** a source language processor.
**
** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
** DEALINGS IN THE SOFTWARE.
** -LICENSE-END-
*/


import Cocoa
import Foundation
import IOKit
import USBControl
import Utility
import PTP

public protocol ConnectionStatusToUIDelegate: AnyObject {
	func onConnection(_ device: USBBulkDevice)
	func onDisconnection(_ device: USBBulkDevice?)
}

open class USBCameraControlController:
			NSViewController,
			USBConnectionManagerDelegate,
			PTPPacketDecoderDelegate {

	let m_usbConnectionManager = USBConnectionManager()
	var m_usbPtpController: USBPTPController?
	public weak var m_connectionStatusDelegate: ConnectionStatusToUIDelegate?
	
	//================================================
	// NSViewController methods
	//================================================
	override open func viewDidLoad() {
		super.viewDidLoad()
		
		m_usbConnectionManager.setDelegate(self)
	}
	
	@discardableResult public func openSession() -> Bool {
		if (m_usbPtpController != nil) {
			return m_usbPtpController!.openSession()
		}
		
		return false
	}
	
	@discardableResult public func stillCapture() -> Bool {
		if (m_usbPtpController != nil) {
			return m_usbPtpController!.stillCapture()
		}
		
		return false
	}
	
	@discardableResult public func startRecord() -> Bool {
		if (m_usbPtpController != nil) {
			return m_usbPtpController!.startRecord()
		}
		
		return false
	}
	
	@discardableResult public func stopRecord() -> Bool {
		if (m_usbPtpController != nil) {
			return m_usbPtpController!.stopRecord()
		}
		
		return false
	}
	
	// All the methods below can be overridden and implemented in a derived ViewController.
	// If the method is not empty, be sure to call the 'super' method from your
	// implementation (eg. 'super.connectedToUSBDevice(device)' )
	
	//================================================
	// USBConnectionManagerDelegate methods
	//================================================
	open func updateDiscoveredUSBPTPDevices(_ discoveredUSBPTPDevices: [USBBulkDevice]) {
		if (!discoveredUSBPTPDevices.isEmpty) {
			m_usbConnectionManager.openDevice(device: discoveredUSBPTPDevices[0])
			
			// The basic camera control view will always connect to the first PTP device found
			m_usbPtpController!.openSession(forceClosePreviousSessionIfPresent: true)
		}
	}

	open func connectedToUSBDevice(_ device: USBBulkDevice) {
		m_usbPtpController = USBPTPController(device: device)
		m_usbPtpController?.setPacketDecodedDelegate(self)
		m_connectionStatusDelegate?.onConnection(device)
		// TODO: Create a PTP interface object
		// TODO: Set the PTP interface object to have a delegate of this class to handle callbacks
	}

	open func disconnectedFromUSBDevice() {
		m_usbPtpController = nil
		m_connectionStatusDelegate?.onDisconnection(nil)
		// Go back to previous view?
	}
}
